/* ---------------------------------------------------------------------------------- */
/* st2sco.c */
/* 
  GIORGI Jean-Francois
  Campus de Beaulieu
  35042 Rennes Cedex FRANCE
  tel: 99 36 20 00
  e-mail: giorgi@irisa.fr
*/
/* ---------------------------------------------------------------------------------- */
#include "st.h"
#include <stdio.h>

/* ---------------------------------------------------------------------------------- */
#define TN_NUM 36
#define OCT_LEN 12
int NotePeriod[TN_NUM] = {      /* Tables des frequences SoundTracker */

/* C      C#     D      D#     E      F      F#     G      G#     A      A#     B  */

 0x0358,0x0328,0x02FA,0x02D0,0x02A6,0x0280,0x025C,0x023A,0x021A,0x01FC,0x01E0,0x01C5,
 0x01AC,0x0194,0x017D,0x0168,0x0153,0x0140,0x012E,0x011D,0x010D,0x00FE,0x00F0,0x00E2,
 0x00D6,0x00CA,0x00BE,0x00B4,0x00AA,0x00A0,0x0097,0x008F,0x0087,0x007f,0x0078,0x0071

};
#define MAX_PER 0x0358+1
int PeriodNote[MAX_PER]; /* reverse of NotePeriod */
#define TN_FIRST 36
/* ---------------------------------------------------------------------------------- */
#define ConvTempo(t) ((3600.0 / (float)(t))*8192.0/10000.0)
/* ---------------------------------------------------------------------------------- */
ST_Score *sts;
ST_Note  sts_patterns[MAX_PAT][PAT_LEN][NUM_VOICES];
BYTE *(Samples[MAX_INST]);
char outfn[300];
char *calloc();
/* ---------------------------------------------------------------------------------- */
CalcPeriodNote()
{
  int i;
  for (i=0;i<MAX_PER;i++) PeriodNote[i] = -1;
  for (i=0;i<TN_NUM;i++) PeriodNote[NotePeriod[i]] = i+TN_FIRST;
}
/* ---------------------------------------------------------------------------------- */
ULONG UpPow2(n)
ULONG n;
{
  ULONG i,spow2;
  for(i=0,spow2=n;spow2>0;i++) spow2 = spow2>>1;
    if ((1<<(i-1)) == n)
      spow2 = n;
    else
      spow2 = 1<<i;
  return(spow2);
}
/* ---------------------------------------------------------------------------------- */
WriteSoundIn(fout,buf,szBuf,szout)
FILE *fout;
BYTE *buf;
ULONG szBuf;
ULONG szout;
{

  BYTE out;
  int n0;
  short sout;

  for(n0=0;n0<szBuf;n0++) {
    out = buf[n0];
    sout = 256*(short)out;
    fwrite(&sout,sizeof(short),1,fout);
  }
  if (szout > szBuf) {
    for(n0=0;n0<szout-szBuf;n0++) {
      out = buf[n0];
      sout = 256*(short)out;
      fwrite(&sout,sizeof(short),1,fout);
    }
  }
}
/* ---------------------------------------------------------------------------------- */
read_module(modfilename)
char *modfilename;
{
  FILE *inf,*fout;
  int i;
  int numpat;
  int ss,sr,srs;
  sts = (ST_Score *)calloc(SZ_ST_SCORE,1);
  inf = fopen(modfilename,"r");
  fread(sts,SZ_ST_SCORE,1,inf);
  fprintf(stderr,"score name : %s\n",sts->sts_name);
/*
  fprintf(stderr,"sts_length : %x\n",(int)sts->sts_length);
  fprintf(stderr,"sts_mark : %x\n",(int)sts->sts_mark);
  fprintf(stderr,"sts_ident : %lx\n",sts->sts_ident);
*/
#ifndef STV1
  if (sts->sts_ident != IDENT_MARK) { 
    fprintf(stderr,"error in module : bad sts_ident\n");
    exit();
  }
#endif

  numpat = 0;
  for (i=0;i<MAX_LENGTH;i++) { /* find greater pattern number */
    if (sts->sts_patpos[i]>numpat) numpat = sts->sts_patpos[i];
  }

  numpat++; /* num of pat. = greater pat. number + 1 */
  fprintf(stderr,"NUMPAT = %d\n",numpat);

  /* sts_patterns = (ST_Note *)calloc(sizeof(ST_Pattern),numpat); */
  fread(&(sts_patterns[0][0][0]),sizeof(ST_Pattern),numpat,inf);

  for (i=0;i<MAX_INST;i++) {
    if (sts->sts_instr[i].sti_midsize != 0) {
      ss = (int)sts->sts_instr[i].sti_midsize;
      sr = (int)sts->sts_instr[i].sti_midrepeatlength;
      srs = (int)sts->sts_instr[i].sti_repeat;
      ss *= 2;
      sr *= 2;
      srs *= 2;
      fprintf(stderr,"sts_instr[%d].sti_name : %s (%d,%d,%d)\n",i,sts->sts_instr[i].sti_name,ss,sr,srs);
      Samples[i] = (BYTE *)calloc(ss,1);
      fread(Samples[i],ss,1,inf);
      sprintf(outfn,"soundin.%d",2*i+1);
      fout = fopen(outfn,"w");
      WriteSoundIn(fout,Samples[i],ss,ss);
      fclose(fout);
      if (sr != DEFAULT_MDRL*2) {
        sprintf(outfn,"soundin.%d",2*i+2);
        fout = fopen(outfn,"w");
        WriteSoundIn(fout,Samples[i]+srs,sr,UpPow2(sr));
        fclose(fout);
      }
    }
  }
  fclose(inf);
}
/* ---------------------------------------------------------------------------------- */
gen_sco_file(rootname)
char *rootname;
{
  FILE *sf;
  int i;
  ULONG ss,sr;
  int t,c,v,ic,pat,endt,sv;
  int oct,pitch;
  int vi[NUM_VOICES];
  int vv[NUM_VOICES];
  int ln[NUM_VOICES];
  ST_Note stnote;
  char scofn[300];
  int tempo;
  float pos,ndur;
  BOOL tpchg;

  strcpy(scofn,rootname);
  strcat(scofn,".sco");

  sf = fopen(scofn,"w");

  fprintf(sf,"; Score Name : %s\n",sts->sts_name);
  for (i=0;i<MAX_INST;i++) {
    ss = (ULONG)sts->sts_instr[i].sti_midsize;
    sr = (ULONG)sts->sts_instr[i].sti_midrepeatlength;
    ss *= 2; sr *= 2; 
    if (ss != 0) {
      fprintf(sf,"f%d 0 %ld 1 %d 0\n",2*i+1,UpPow2(ss),2*i+1);
      if (sr != DEFAULT_MDRL*2) {
        fprintf(sf,"f%d 0 %ld 1 %d 0\n",2*i+2,UpPow2(sr),2*i+2);
      }
    }
  }

  tempo = 6;
  tpchg = FALSE;
  fprintf(sf,"t 0 %f\n",60.0*8192.0/10000.0);
  endt = (int)sts->sts_length; 
  pos = 1.0;
  ndur = tempo/60.0;
  for (v=0;v<NUM_VOICES;v++) {vi[v] = 0; ln[v] = 0;vv[v] = MAX_VOLUME;}
  for(t=0;t<endt;t++) {
    pat = sts->sts_patpos[t];
    for (c=0;c<PAT_LEN;c++,pos+=ndur) {
      for (v=0;v<NUM_VOICES;v++) {
        stnote = sts_patterns[pat][c][v]; /* read event */

        if (stnote.stn_period != 0 && ln[v] != 0 && PeriodNote[ln[v]] != -1) {
          /* new note,so stop previous one */
	  oct   = PeriodNote[ln[v]]/OCT_LEN;
	  pitch = PeriodNote[ln[v]]%OCT_LEN;
	  fprintf(sf,"i -%d.0%d %f -%f %d %d %d %d\n",vi[v],v,pos,ndur,ln[v],vv[v],0,tempo);
        }
        ic = (int)stnote.stn_ins + (int)(stnote.stn_upperins)*16;
        if (ic!=0) {
          vi[v] = ic;
        }

	vv[v] = ic != 0 ? (sts->sts_instr[ic-1].sti_volume) : -1;
        if (stnote.stn_period != 0) {
	  ln[v] = stnote.stn_period;
        }
	if (stnote.stn_ctrl == STC_VOLUME)
	    vv[v] = stnote.stn_data;
	if (stnote.stn_ctrl == STC_SLIDEVOL) {
	  sv = stnote.stn_data & 0x0f;
	  if (stnote.stn_data & 0xf0 != 0)
	    sv = -sv;
	}
	else {
	  sv = 0;
	}
        if (ln[v] != 0 && PeriodNote[ln[v]] != -1) {
	  oct   = PeriodNote[ln[v]]/OCT_LEN;
	  pitch = PeriodNote[ln[v]]%OCT_LEN;
          if (vi[v] != 0)
	    fprintf(sf,"i %d.0%d %f -%f %d %d %d %d\n",vi[v],v,pos,ndur,ln[v],vv[v],sv,tempo);
        }
	switch (stnote.stn_ctrl) {
	  case STC_JUMP :
	    if (stnote.stn_data>t) { /* only forward jump */
              v=NUM_VOICES;c=PAT_LEN; t=stnote.stn_data; break;
            }
	    /* else do a BREAK */
	  case STC_BREAK :
	    v=NUM_VOICES;c=PAT_LEN;break;
	  case STC_SPEED :
	    if (stnote.stn_data > 0xF)
	      fprintf(stderr,"warning: STC_SPEED greater than 0x0f\n");
            if (tempo != stnote.stn_data&0x0F) {
              tempo = stnote.stn_data&0x0F;
	      tpchg = TRUE;
            }
	    break;
	} /* switch */
      } /* for v */
      if (tpchg == TRUE) {
        ndur = tempo / 60.0;
	tpchg = FALSE;
      }
    } /* for c */
  } /* for t */
  fprintf(sf,"e\n");
  fclose(sf);
}
/* ---------------------------------------------------------------------------------- */
AddCommon(of)
FILE *of;
{
  fprintf(of,"tigoto suite\n");
  fprintf(of,"is = (int(p1)-1)*2+1\n");
  fprintf(of,"ifreq = 0.279365 * p4\n");
  fprintf(of,"ifreq = 1000000/ifreq\n");
  fprintf(of,"ionedur = ioneshot/ifreq \n");
  fprintf(of,"irepdur = irepeat/ifreq \n");
  fprintf(of,"isamp = p5\n");
  fprintf(of,"ieamp = isamp\n");
  fprintf(of,"suite:\n");
  fprintf(of,"isamp = ieamp\n");
  fprintf(of,"if p5 == -1 igoto volcont\n");
  fprintf(of,"isamp = p5\n");
  fprintf(of,"ieamp = isamp\n");
  fprintf(of,"volcont:\n");
  fprintf(of,"ieamp = ieamp + p6*p7\n");
  fprintf(of,"if ieamp > 0 igoto ampcont1\n");
  fprintf(of,"ieamp = 0\n");
  fprintf(of,"ampcont1:\n");
  fprintf(of,"if ieamp < 64 igoto ampcont2\n");
  fprintf(of,"ieamp = 64\n");
  fprintf(of,"ampcont2:\n");
  fprintf(of,"aamp    line isamp,abs(p3),ieamp\n");
  fprintf(of,"tigoto suite2\n");
  fprintf(of,"aone    line 0,ionedur,ioneshot\n");
  fprintf(of,"aout    tablei aone,is,0,0,0\n");
  fprintf(of,"        out     aout*aamp*givol\n");
  fprintf(of,"if irepeat == 2 goto suite2\n");
  fprintf(of,"arep    line ionedur,irepdur,irepeat\n");
  fprintf(of,"arout   tablei arep,is+1,0,0,1\n");
  fprintf(of,"        out     arout*aamp*givol\n");
  fprintf(of,"suite2:\n");
  fprintf(of,"\n");
}
/* ---------------------------------------------------------------------------------- */
gen_orc_file(rootname)
char *rootname;
{
  FILE *of;
  int i;
  char orcfn[300];
  ULONG ss;

  strcpy(orcfn,rootname);
  strcat(orcfn,".orc");

  of = fopen(orcfn,"w");
  fprintf(of,"; Orchestrafile for %s.sco\n",rootname);
  fprintf(of,"sr=8192\nkr=512\nksmps=16\n");


  fprintf(of,"\ngivol   init	3*64 ;global volume\n");

  for (i=0;i<MAX_INST;i++) {
    ss = (ULONG)sts->sts_instr[i].sti_midsize;
    ss *= 2;
    if (ss!=0) {
      fprintf(of,";--------------------------------------------\n");
      fprintf(of,"instr %d		;%s\n",i+1,sts->sts_instr[i].sti_name);
      fprintf(of,"ioneshot = %d\n",ss);
      fprintf(of,"irepstart = %d\n",sts->sts_instr[i].sti_repeat * 2);
      fprintf(of,"irepeat= %d\n",sts->sts_instr[i].sti_midrepeatlength * 2);
/*      fprintf(of,"ivolume = %d\n",sts->sts_instr[i].sti_volume); */
      AddCommon(of);
      fprintf(of,"endin\n");
    }
  }
  fprintf(of,";--------------------------------------------\n");
  fclose(of);
}
/* ---------------------------------------------------------------------------------- */
main(argc,argv)
int argc;
char **argv;
{
  if (argc != 3) {
     fprintf(stderr,"usage : %s st-module csound-root-filename\n",argv[0]);
     exit();
  }
  read_module(argv[1]);
  CalcPeriodNote();
  gen_sco_file(argv[2]);
  gen_orc_file(argv[2]);
}
/* ---------------------------------------------------------------------------------- */
